/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.lib.daten;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.List;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;

/**
 * Klasse zum Test der Funktionalität einer Attributliste.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class AttributlisteTest {

	/** Regel zur Auswertung erwarteter Exceptions. */
	@Rule
	public ExpectedException thrown = ExpectedException.none();

	/**
	 * Testliste mit Attributen.
	 */
	private List<Attribut<?>> attribute;

	/**
	 * zweite Testliste mit Attributen.
	 */
	private List<Attribut<?>> attribute2;

	/**
	 * Füllen der Attributlisten zum Test.
	 */
	@Before
	public void attributeAnlegen() {
		attribute = new ArrayList<>();
		attribute.add(new Kommazahl("Kommazahl", 13.657));
		attribute.add(new Ganzzahl("Ganzzahl", 678L));
		attribute.add(new Zeichenkette("Zeichenkette", "Heinz"));

		attribute2 = new ArrayList<>();
		attribute2.add(new Kommazahl("Kommazahl", 99.9));
		attribute2.add(new Ganzzahl("Ganzzahl", 999L));
		attribute2.add(new Zeichenkette("Zeichenkette", "Willi"));
	}

	/**
	 * Test: Hinzufügen zur Liste.
	 */
	@Test
	public void testAdd() {
		final Attributliste liste = new Attributliste("DieListe");
		Assert.assertFalse(liste.hasAttribut("Zahl"));
		liste.add(new Kommazahl("Zahl"));
		Assert.assertTrue(liste.hasAttribut("Zahl"));
	}

	/**
	 * Test des Konstruktors mit Angabe einer Zwichenkette.
	 */
	@Test
	public void testAttributlisteString() {
		final Attributliste liste = new Attributliste("DieListe");
		Assert.assertFalse(liste.hasAttribut("Zahl"));
	}

	/**
	 * Konstruktortest mit einer Liste als Argument.
	 */
	@Test
	public void testAttributlisteStringListOfAttributOfQ() {
		final Attributliste liste = new Attributliste("DieListe", attribute);
		Assert.assertFalse(liste.hasAttribut("Zahl"));
		Assert.assertTrue(liste.hasAttribut("Kommazahl"));
		Assert.assertTrue(liste.hasAttribut("Ganzzahl"));
		Assert.assertTrue(liste.hasAttribut("Zeichenkette"));
	}

	/**
	 * Test: Liste auslesen.
	 */
	@SuppressWarnings("cast")
	@Test
	public void testGet() {
		final Attributliste liste = new Attributliste("DieListe", attribute);
		Assert.assertTrue(liste.get().size() == 3);
		Assert.assertTrue(liste.get() instanceof List);
	}

	/**
	 * Test: Attribut auslesen.
	 */
	@Test
	public void testGetAttribut() {
		final Attributliste liste = new Attributliste("DieListe", attribute);

		Assert.assertEquals(13.657, liste.getAttribut("Kommazahl").get());
		Assert.assertEquals(678L, liste.getAttribut("Ganzzahl").get());
		Assert.assertEquals("Heinz", liste.getAttribut("Zeichenkette").get());
	}

	/**
	 * Test: Attribut auslesen.
	 */
	@Test
	public void testGetUndefinedAttribut() {
		final Attributliste liste = new Attributliste("DieListe", attribute);

		thrown.expect(IllegalArgumentException.class);
		Assert.assertEquals(null, liste.getAttribut("Willi"));
	}

	/**
	 * Test: Namen der Liste auslesen.
	 */
	@Test
	public void testGetName() {
		final Attributliste liste = new Attributliste("DieListe", attribute);
		Assert.assertEquals("DieListe", liste.getName());
	}

	/**
	 * Test: Serialisierung und Deserialisierung der Liste.
	 *
	 * @throws IOException
	 *             Fehler beim Schreiben der Daten
	 * @throws ClassNotFoundException
	 *             beim Deserialisieren wurde eine Klasse nicht gefunden
	 */
	@Test
	public void testSerialization() throws IOException, ClassNotFoundException {
		Attributliste liste = new Attributliste("DieListe", attribute);
		liste.add(new Attributliste("Sub", attribute2));

		final ByteArrayOutputStream byteOut = new ByteArrayOutputStream(1000);
		final ObjectOutputStream output = new ObjectOutputStream(byteOut);
		output.writeObject(liste);
		final byte[] bytes = byteOut.toByteArray();

		final ObjectInputStream input = new ObjectInputStream(new ByteArrayInputStream(bytes));
		liste = (Attributliste) input.readObject();

		Assert.assertEquals(13.657, liste.getAttribut("Kommazahl").get());
		Assert.assertEquals(678L, liste.getAttribut("Ganzzahl").get());
		Assert.assertEquals("Heinz", liste.getAttribut("Zeichenkette").get());

		liste = (Attributliste) liste.getAttribut("Sub");
		Assert.assertEquals(99.9, liste.getAttribut("Kommazahl").get());
		Assert.assertEquals(999L, liste.getAttribut("Ganzzahl").get());
		Assert.assertEquals("Willi", liste.getAttribut("Zeichenkette").get());

	}

	/**
	 * Test; Attribute einer Liste setzen.
	 */
	@Test
	public void testSet() {
		final Attributliste liste = new Attributliste("DieListe", attribute);
		liste.set(attribute2);
		Assert.assertEquals(99.9, liste.getAttribut("Kommazahl").get());
		Assert.assertEquals(999L, liste.getAttribut("Ganzzahl").get());
		Assert.assertEquals("Willi", liste.getAttribut("Zeichenkette").get());
	}
}
