/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.param;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.Blob;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.text.DateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;

import javax.sql.rowset.serial.SerialBlob;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.puk.param.lib.DatenSchreibenException;
import de.bsvrz.puk.param.lib.Parameter;
import de.bsvrz.puk.param.lib.ParameterInfo;
import de.bsvrz.puk.param.lib.daten.Attribut;
import de.bsvrz.puk.param.lib.daten.DataWithTime;
import de.bsvrz.puk.param.lib.daten.Datensatz;
import de.bsvrz.puk.param.lib.daten.ParameterDatenFactory;
import de.bsvrz.puk.param.lib.daten.ParameterDatenFactory.DataWithState;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Persistenzmodul zum Speichern der Daten der Parametrierung unter Verwendung
 * einer eingebauten Apache-Derby-Datenbank.
 *
 * @author BitCtrl Systems GmbH, Uwe Peuker
 */
class DerbyPersistenz implements Persistenz {

	/** ein leeres Byte-Array. */
	private static final byte[] EMPTY_BLOB = new byte[0];

	/**
	 * Definition für den Schlüssel auf Datensätze innerhalb der
	 * Derby-Datenbank.
	 */
	private static final class DbSpezifikation {

		/** die PID des zugeordneten Objekts. */
		private final String pid;

		/** die ID der zugeordneten Attributgruppe. */
		private final long atg;

		/** die zugeordnete Simulationsvariante. */
		private final short sim;

		/**
		 * Konstruktor.
		 *
		 * @param pid
		 *            die Objekt-PID
		 * @param atg
		 *            die ID der Attributgruppe
		 * @param sim
		 *            die Simulationsvariante
		 */
		private DbSpezifikation(final String pid, final long atg, final short sim) {
			this.pid = pid;
			this.atg = atg;
			this.sim = sim;
		}
	}

	/**
	 * Cache zum zwischenspeichern von Datensätzen, um den Zugriff auf die
	 * externen Medien zu optimieren.
	 */
	private static class PersistenzCache {

		/**
		 * Eintrag innerhalb des Caches des Persitenzmoduls.
		 */
		private static final class CacheData {

			/** die gespeicherten Daten. */
			private final DataWithTime datenObjekt;

			/** markiert, ob die Datenbank gültige Daten enthält. */
			private final boolean datenVerfuegbar;

			/**
			 * Konstruktor.
			 *
			 * @param verfuegbarkeit
			 *            die Verfügbarkeit von Daten in der Datenbank des
			 *            Persistenzmoduls
			 */
			private CacheData(final boolean verfuegbarkeit) {
				this.datenObjekt = null;
				this.datenVerfuegbar = verfuegbarkeit;
			}

			/**
			 * Konstruktor.
			 *
			 * @param daten
			 *            die Daten
			 */
			private CacheData(final DataWithTime daten) {
				this.datenObjekt = daten;
				this.datenVerfuegbar = daten != null;
			}
		}

		/**
		 * Schlüssel zum Zugriff auf den Persitenzcache.
		 */
		private static class CacheKey {

			/** die ID eines Systemsobjekts. */
			private final long objId;

			/** die ID eines Attributgruppe. */
			private final long atgId;

			/** die Simulationsvariante. */
			private final short simVariante;

			/**
			 * Konstruktor.
			 *
			 * @param objId
			 *            Objekt-ID
			 * @param atgId
			 *            ATG-ID
			 * @param simVariante
			 *            die Simulationsvariante
			 */
			CacheKey(final long objId, final long atgId, final short simVariante) {
				this.objId = objId;
				this.atgId = atgId;
				this.simVariante = simVariante;
			}

			@Override
			public boolean equals(final Object obj) {
				boolean result = false;
				if (this == obj) {
					result = true;
				} else {
					if (obj instanceof CacheKey) {
						final CacheKey other = (CacheKey) obj;
						if (atgId == other.atgId && objId == other.objId && simVariante == other.simVariante) {
							result = true;
						}
					}
				}
				return result;
			}

			@Override
			public int hashCode() {
				return (int) objId;
			}

			@Override
			public String toString() {
				return "{" + objId + ";" + atgId + ";" + simVariante + "}";
			}
		}

		/** maximale Anzahl der im internen Cache aufbewahrten Datensätze. */
		private int maximaleCacheAnzahl = 1;

		/** die intern zwischengespeicherten Daten. */
		private final Map<CacheKey, CacheData> daten = new HashMap<>();

		/**
		 * Set mit den Keys, über die im Moment Daten im Cache abfragbar sind.
		 */
		private final Set<CacheKey> dataKeys = new LinkedHashSet<>();

		/** die Persistenz, deren Daten gecachet werden. */
		private final DerbyPersistenz persistenz;

		/**
		 * Standardkonstruktor.
		 *
		 * @param persistenz
		 *            das Persistenzmodul
		 * @param anzahl
		 *            die maximale Größe des internen Datencaches
		 */
		PersistenzCache(final DerbyPersistenz persistenz, final int anzahl) {
			this.persistenz = persistenz;
			this.maximaleCacheAnzahl = anzahl;
		}

		/**
		 * fügt der Menge der Schlüssel, für die Daten im Cache zu finden sind
		 * einen Eintrag hinzu.
		 *
		 * @param key
		 *            der Schlüssel.
		 */
		private void addKey(final CacheKey key) {
			synchronized (daten) {
				dataKeys.remove(key);
				dataKeys.add(key);
				int size = dataKeys.size() - maximaleCacheAnzahl;
				if (size > 0) {
					final Collection<CacheKey> entfernbar = new ArrayList<>();
					final Iterator<CacheKey> iterator = dataKeys.iterator();
					while (size > 0 && iterator.hasNext()) {
						entfernbar.add(iterator.next());
						size--;
					}
					for (final CacheKey cacheKey : entfernbar) {
						final CacheData alteDaten = daten.remove(cacheKey);
						boolean verfuegbarkeit = false;
						if (alteDaten != null) {
							verfuegbarkeit = alteDaten.datenVerfuegbar;
						}
						daten.put(cacheKey, new CacheData(verfuegbarkeit));
						dataKeys.remove(cacheKey);
					}
				}
			}
		}

		/**
		 * ermittelt, ob Daten für die angegebene Spezifikation bereits
		 * abgerufen wurden.
		 *
		 * @param objId
		 *            die Objekt-ID
		 * @param atgId
		 *            die ATG-ID
		 * @param simVariante
		 *            die Simulationsvariante.
		 * @return <code>true</code>, wenn Daten im Cache vorhanden sind.
		 */
		boolean contains(final long objId, final long atgId, final short simVariante) {
			return daten.containsKey(new CacheKey(objId, atgId, simVariante));
		}

		/**
		 * liest die Daten für die angegebene Spezifikation.
		 *
		 * @param obj
		 *            das Objekt zu dem die Daten gehören
		 * @param atg
		 *            die Attributgruppe, in der die daten abgelegt sind
		 * @param simVariante
		 *            die Simulationsvariante.
		 * @return der Datensatz oder <code>null</code>, wenn kein Datensatz
		 *         verfügbar ist.
		 */
		DataWithTime get(final SystemObject obj, final AttributeGroup atg, final short simVariante) {
			final CacheKey key = new CacheKey(obj.getId(), atg.getId(), simVariante);
			CacheData datenEintrag = daten.get(key);
			DataWithTime result = null;
			if (datenEintrag != null && datenEintrag.datenVerfuegbar) {
				if (datenEintrag.datenObjekt == null) {
					datenEintrag = new CacheData(persistenz.liesIntern(obj, atg, simVariante));
					daten.put(key, datenEintrag);
					if (datenEintrag.datenObjekt != null) {
						addKey(key);
					}
				}
				result = datenEintrag.datenObjekt;
			}
			return result;
		}

		/**
		 * trägt die Daten für die angegebene Spezifikation ein.
		 *
		 * @param objId
		 *            die Objekt-ID
		 * @param atgId
		 *            die ATG-ID
		 * @param simVariante
		 *            die Simulationsvariante.
		 * @param datenSatz
		 *            die Daten
		 */
		void put(final long objId, final long atgId, final short simVariante, final DataWithTime datenSatz) {
			synchronized (daten) {
				final CacheKey key = new CacheKey(objId, atgId, simVariante);
				if (datenSatz == null) {
					final CacheData oldData = daten.put(key, new CacheData(datenSatz));
					if (oldData != null && oldData.datenObjekt != null) {
						dataKeys.remove(key);
					}
				} else {
					final CacheData oldData = daten.put(key, new CacheData(datenSatz));
					if (oldData == null || oldData.datenObjekt == null) {
						addKey(key);
					}
				}
			}
		}

		/**
		 * entfernt die definierten Daten aus dem Cache.
		 *
		 * @param objId
		 *            die Objekt-ID
		 * @param atgId
		 *            die ATG-ID
		 * @param simVariante
		 *            die Simulationsvariante.
		 */
		void remove(final long objId, final long atgId, final short simVariante) {
			synchronized (daten) {
				final CacheKey key = new CacheKey(objId, atgId, simVariante);
				final CacheData datenSatz = daten.remove(key);
				if (datenSatz != null && datenSatz.datenObjekt != null) {
					dataKeys.remove(key);
				}
			}
		}

		/**
		 * entfernt alle Daten einer Simulationsvariante.
		 *
		 * @param sim
		 *            die Simulationsvariante
		 */
		void remove(final short sim) {
			synchronized (daten) {
				final Collection<CacheKey> zuEntfernen = new ArrayList<>();
				for (final CacheKey key : daten.keySet()) {
					if (key.simVariante == sim) {
						zuEntfernen.add(key);
					}
				}

				for (final CacheKey key : zuEntfernen) {
					final CacheData datenSatz = daten.remove(key);
					if (datenSatz != null && datenSatz.datenObjekt != null) {
						dataKeys.remove(key);
					}
				}
			}
		}

		/**
		 * liefert die Anzahl der Cache-Einträge.
		 *
		 * @return die Grösse
		 */
		long size() {
			return daten.size();
		}
	}

	/** Liste der angemeldeten Simulationen. */
	private static Set<Short> simulationen = new HashSet<>();

	/** der Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/** Speicher für veröffentlichte Daten. */
	private PersistenzCache cache;

	/** die Datenbankverbindung zur Derby-Datenbank. */
	private Connection dbConnection;

	/** Lockobjekt zur Synchronisation der Datenbankzugriffe. */
	private final Object lock = new Object();

	/**
	 * SQL-Statement zum Löschen aller Daten einer Simulationsvariante aus der
	 * Datenbank.
	 */
	// private PreparedStatement simRemoveStatement;

	/** die verwendete Datenverteilerverbindung. */
	private ClientDavInterface dav;

	/**
	 * Konstruktor.
	 */
	DerbyPersistenz() {
		DerbyPersistenz.LOGGER.info("Starte Submodul DerbyPersistenz ...");
	}

	/**
	 * aktualisiert den übergebenen Datensatz in die Datenbank.
	 *
	 * @param obj
	 *            das Objekt
	 * @param atgId
	 *            die Id der Attributgruppe
	 * @param sim
	 *            die Simulationsvariante
	 * @param daten
	 *            der Datensatz
	 * @throws DatenSchreibenException
	 *             die Daten konnten nicht innerhalb der Datenbank gespeichert
	 *             werden
	 */
	private void aktualisiereIntern(final SystemObject obj, final long atgId, final short sim, final DataWithTime daten)
			throws DatenSchreibenException {

		final Datensatz datenSatz = (Datensatz) ParameterDatenFactory.getSerializable(dav, daten.getData());

		synchronized (lock) {

			try (PreparedStatement updateStatement = dbConnection.prepareStatement(
					"UPDATE parameterdaten SET daten=?, zeitpunkt=? WHERE pid=? AND atg=? AND simvariante=?")) {

				try (Statement s = dbConnection.createStatement()) {
					s.execute("LOCK TABLE PARAMETERDATEN IN EXCLUSIVE MODE");

					final Blob blob = getBlob(datenSatz);
					updateStatement.setBlob(1, blob);
					updateStatement.setLong(2, daten.getZeitpunkt());
					updateStatement.setString(3, obj.getPid());
					updateStatement.setLong(4, atgId);
					updateStatement.setInt(5, sim);
					updateStatement.execute();
				}

			} catch (final SQLException e) {
				LogTools.log(DerbyPersistenz.LOGGER, new BetriebsmeldungDaten(obj),
						ParamMessage.ErrorSchreibenDatensatzDerbyDb, obj, dav.getDataModel().getObject(atgId), sim,
						e.getMessage());
				reconnect();
				throw new DatenSchreibenException("Fehler beim Aktualisieren der Daten für: " + obj.getPid(), e);
			} catch (final IOException e) {
				LogTools.log(DerbyPersistenz.LOGGER, ParamMessage.ErrorSchreibenDatensatzDerbyDb, obj,
						dav.getDataModel().getObject(atgId), sim, e.getMessage());
				throw new DatenSchreibenException("Fehler beim Aktualisieren der Daten für: " + obj.getPid(), e);
			}
		}
	}

	@Override
	public void aufraeumen() {
		synchronized (lock) {
			try {
				reconnect();

				try (CallableStatement bereinigungsStatement = dbConnection.prepareCall(
						"call SYSCS_UTIL.SYSCS_INPLACE_COMPRESS_TABLE('PARAMETRIERUNG', 'PARAMETERDATEN', 1, 1, 1)")) {
					try (Statement s = dbConnection.createStatement()) {

						s.execute("LOCK TABLE PARAMETERDATEN IN EXCLUSIVE MODE");

						DerbyPersistenz.LOGGER.info("Bereinige Derby-Datenbank");
						bereinigungsStatement.execute();
						DerbyPersistenz.LOGGER.info("Derby-Datenbank bereinigt");
					}
				}

				close();
				connect(PersistenzParameter.getInstanz());

			} catch (final SQLException e) {
				DerbyPersistenz.LOGGER
						.error("Bereinigung der Derby-DB konnte nicht ausgeführt werden: " + e.getMessage(), e);
				reconnect();
			} catch (final ClassNotFoundException e) {
				DerbyPersistenz.LOGGER.error("Derby-Modul konnte nicht initialisiert werden", e);
			}
		}
	}

	@Override
	public void backup() {
		synchronized (lock) {
			try {
				reconnect();

				DerbyPersistenz.LOGGER.info("Starte Backup der Parameterdatenbank!");

				final DateFormat todaysDate = new java.text.SimpleDateFormat("yyyy-MM-dd-HH-mm");
				final String backupdirectory = PersistenzParameter.getInstanz().getPersistenzWurzel() + "/backup/"
						+ todaysDate.format(java.util.Calendar.getInstance().getTime());

				try (CallableStatement cs = dbConnection.prepareCall("CALL SYSCS_UTIL.SYSCS_BACKUP_DATABASE(?)")) {
					cs.setString(1, backupdirectory);
					cs.execute();
				}
				DerbyPersistenz.LOGGER.info("Backup in " + backupdirectory + " erstellt!");
				close();
				connect(PersistenzParameter.getInstanz());

			} catch (final SQLException e) {
				DerbyPersistenz.LOGGER.error("Backup der Derby-DB konnte nicht ausgeführt werden: " + e.getMessage(),
						e);
				reconnect();
			} catch (final ClassNotFoundException e) {
				DerbyPersistenz.LOGGER.error("Derby-Modul konnte nicht initialisiert werden", e);
			}
		}
	}

	/**
	 * Die Funktion speichert alle ausstehenden Datenbankänderungen und schließt
	 * die Datenbankverbindung.
	 */
	@Override
	public void close() {
		try {
			dbConnection.commit();
			dbConnection.close();
		} catch (final SQLException e) {
			LogTools.log(DerbyPersistenz.LOGGER, ParamMessage.ErrorCloseDerbyDb, e.getMessage());
		}
	}

	@Override
	public void entferne(final ParameterInfo parameter) {
		synchronized (cache) {
			cache.remove(parameter.getObjekt().getId(), parameter.getAtg().getId(), parameter.getSim());
		}
		entferneIntern(parameter.getObjekt().getPid(), parameter.getAtg().getId(), parameter.getSim());
	}

	/**
	 * entfernt den Eintrag mit der übergebenen Datenspezifikation aus der
	 * Datenbank.
	 *
	 * @param pid
	 *            die PID des Objekts
	 * @param id
	 *            die ID der Attributgruppe
	 * @param sim
	 *            die Simulationsvariante
	 */
	private void entferneIntern(final String pid, final long id, final short sim) {

		synchronized (lock) {
			try (PreparedStatement removeStatement = dbConnection
					.prepareStatement("DELETE FROM parameterdaten WHERE pid=? AND atg=? AND simvariante=?")) {

				try (Statement s = dbConnection.createStatement()) {
					s.execute("LOCK TABLE PARAMETERDATEN IN EXCLUSIVE MODE");

					removeStatement.setString(1, pid);
					removeStatement.setLong(2, id);
					removeStatement.setInt(3, sim);
					removeStatement.execute();
				}
			} catch (final SQLException e) {
				LogTools.log(DerbyPersistenz.LOGGER, ParamMessage.ErrorLoescheDatensatzDerbyDb,
						dav.getDataModel().getObject(pid), dav.getDataModel().getObject(id), sim, e.getMessage());
				reconnect();
			}
		}
	}

	/** stellt die Datenbankverbindung nach einem Fehler wieder her. */
	private void reconnect() {
		try {
			if (!dbConnection.isValid(0)) {
				close();
				connect(PersistenzParameter.getInstanz());
			}
		} catch (final SQLException e) {
			DerbyPersistenz.LOGGER.error(e.getLocalizedMessage(), e);
		} catch (final ClassNotFoundException e) {
			DerbyPersistenz.LOGGER.error(e.getLocalizedMessage(), e);
		}
	}

	@Override
	public void entferneSimulation(final short simulationsVariante) {
		DerbyPersistenz.LOGGER.info("Entferne Daten für Simulationsvariante: " + simulationsVariante);

		synchronized (lock) {

			try (PreparedStatement simRemoveStatement = dbConnection
					.prepareStatement("DELETE FROM parameterdaten WHERE simvariante=?")) {

				try (Statement s = dbConnection.createStatement()) {
					s.execute("LOCK TABLE PARAMETERDATEN IN EXCLUSIVE MODE");

					simRemoveStatement.setInt(1, simulationsVariante);
					simRemoveStatement.execute();
				}
			} catch (final SQLException e) {
				LogTools.log(DerbyPersistenz.LOGGER, ParamMessage.ErrorLoescheSimVarianteDerbyDb, simulationsVariante,
						e.getMessage());
				reconnect();
			}
		}
		synchronized (cache) {
			cache.remove(simulationsVariante);
		}
		DerbyPersistenz.simulationen.remove(simulationsVariante);
		DerbyPersistenz.LOGGER.info("Daten für Simulationsvariante: " + simulationsVariante + " entfernt");
	}

	/**
	 * erzeugt die Tabelle zum Speichern der Daten in der Parameterdatenbank.
	 *
	 * @throws SQLException
	 *             Fehler beim Anlegen der Tabelle aufgetreten
	 */
	private void erzeugeDatenTabelle() throws SQLException {

		boolean tabelleExistiert = false;

		try (ResultSet set = dbConnection.getMetaData().getTables(null, null, "PARAMETERDATEN", null)) {
			while (set.next()) {
				if ("PARAMETERDATEN".equalsIgnoreCase(set.getString("TABLE_NAME"))) {
					tabelleExistiert = true;
					break;
				}
			}
		}

		if (!tabelleExistiert) {
			try (Statement s = dbConnection.createStatement()) {
				s.execute(
						"create table parameterdaten(pid varchar(512), atg bigint, simvariante smallint, zeitpunkt bigint, daten blob, PRIMARY KEY ( pid, atg, simvariante))");
			}
		}

		boolean zeitSpalteExists = false;

		try (ResultSet set = dbConnection.getMetaData().getColumns(null, null, "PARAMETERDATEN", null)) {
			while (set.next()) {
				if ("ZEITPUNKT".equalsIgnoreCase(set.getString("COLUMN_NAME"))) {
					zeitSpalteExists = true;
					break;
				}
			}
		}
		if (!zeitSpalteExists) {
			try (Statement s = dbConnection.createStatement()) {
				s.execute("alter table parameterdaten add zeitpunkt bigint default 0");
			}
		}
	}

	@Override
	public void erzeugeSimulation(final short simulationsVariante) {
		DerbyPersistenz.LOGGER.info("Registriere Simulationsvariante: " + simulationsVariante);
	}

	/**
	 * Umwandlung der Parameterdaten in ein BLOB-Objekt zur Speicherung in der
	 * Datenbank.
	 *
	 * @param daten
	 *            die zu speichernden Daten
	 * @return das speicherbare BLOB-Objekt
	 * @throws SQLException
	 *             Fehler bei der Umwandlung in die Ziel-Struktur
	 * @throws IOException
	 *             allgemeiner Konvertierungsfehler
	 */
	private Blob getBlob(final Attribut<? extends Object> daten) throws SQLException, IOException {
		SerialBlob result = null;
		if (daten != null) {
			final ByteArrayOutputStream stream = new ByteArrayOutputStream();
			try (ObjectOutputStream ausgabe = new ObjectOutputStream(stream)) {
				ausgabe.writeObject(daten);
				ausgabe.flush();
				result = new SerialBlob(stream.toByteArray());
			}
		} else {
			result = new SerialBlob(DerbyPersistenz.EMPTY_BLOB);
		}
		return result;
	}

	/**
	 * wandelt ein in der Datenbank gespeichertes BLOB-Objekt in die
	 * Datenstruktur des Parameterdatensatzes um.
	 *
	 * @param blob
	 *            das BLOB-Objekt
	 * @return die Datenstruktur
	 */
	private List<Attribut<?>> getDatenSatz(final Blob blob) {
		List<Attribut<?>> result = null;
		if (blob != null) {
			try (ObjectInputStream eingabe = new ObjectInputStream(blob.getBinaryStream())) {
				final Datensatz daten = (Datensatz) eingabe.readObject();
				result = daten.get();
			} catch (final IOException e) {
				LogTools.log(DerbyPersistenz.LOGGER, ParamMessage.ErrorLesenBlobDerbyDb, e.getMessage());
			} catch (final SQLException e) {
				LogTools.log(DerbyPersistenz.LOGGER, ParamMessage.ErrorLesenBlobDerbyDb, e.getMessage());
			} catch (final ClassNotFoundException e) {
				LogTools.log(DerbyPersistenz.LOGGER, ParamMessage.ErrorLesenBlobDerbyDb, e.getMessage());
			}
		}
		return result;
	}

	/**
	 * Die Funktion stellt eine Verbindung zur eingebetteten Derby-Datenbank her
	 * und erzeugt gegebenenfalls die notwendigen Datenstrukturen.
	 *
	 * @param verbindung
	 *            die Datenverteilerverbindung
	 * @param parameter
	 *            die Parameter für das Persistenzmodul
	 */
	@Override
	public void initalisiere(final ClientDavInterface verbindung, final PersistenzParameter parameter) {

		this.dav = verbindung;
		cache = new PersistenzCache(this, parameter.getCacheAnzahl());

		Exception connectionError = null;

		try {
			connect(parameter);

			erzeugeDatenTabelle();

			aufraeumen();
			initialisiereCache();

		} catch (final SQLException e) {
			connectionError = e;
		} catch (final ClassNotFoundException e) {
			connectionError = e;
		}

		if (connectionError != null) {
			String cause = null;
			if (connectionError.getCause() != null) {
				cause = connectionError.getCause().getMessage();
			}
			LogTools.log(DerbyPersistenz.LOGGER, ParamMessage.ErrorInitialisiereDerbyDb, connectionError.getMessage(),
					cause);
			throw new RuntimeException(connectionError);
		}

		DerbyPersistenz.LOGGER
				.info("Initialisiere Datenbankbereinigung mit Zyklus: " + parameter.getBereinigungsZyklus());
		if (parameter.getBereinigungsZyklus() > 0) {
			new Timer("Derby-Cleaner", true).schedule(new TimerTask() {

				@Override
				public void run() {
					aufraeumen();
				}
			}, parameter.getBereinigungsZyklus(), parameter.getBereinigungsZyklus());
		}
	}

	/**
	 * stellt die Verbindung zur Datenbank her.
	 *
	 * @param parameter
	 *            die Parameter des Persistenzmoduls
	 * @throws ClassNotFoundException
	 *             Treiber für die Datenbank wurde nicht gefunden
	 * @throws SQLException
	 *             Fehler beim Datenbankzugriff
	 */
	private void connect(final PersistenzParameter parameter) throws ClassNotFoundException, SQLException {
		final Properties props = new Properties();
		props.put("user", "parametrierung");
		props.put("password", "parametrierung");

		System.setProperty("derby.system.home", parameter.getPersistenzWurzel());
		Class.forName("org.apache.derby.jdbc.EmbeddedDriver");
		dbConnection = DriverManager.getConnection("jdbc:derby:parameterDB;create=true", props);
		dbConnection.setAutoCommit(true);
	}

	/**
	 * lädt die Daten aus der Datenbank in eine interne Liste, um den Zugriff zu
	 * beschleunigen.
	 *
	 * @throws SQLException
	 *             die Daten konnten nicht korrekt eingelesen werden
	 */
	private void initialisiereCache() throws SQLException {

		DerbyPersistenz.LOGGER.info("Initialisiere Parameterdaten");
		final Collection<DbSpezifikation> entfernbareEintrage = new ArrayList<>();

		final List<Parameter> timeUpdates = new ArrayList<>();
		final List<Parameter> dataUpdates = new ArrayList<>();

		try (PreparedStatement stmt = dbConnection
				.prepareStatement("SELECT pid, atg, simvariante, daten, zeitpunkt FROM parameterdaten")) {

			try (ResultSet set = stmt.executeQuery()) {

				while (set.next()) {
					final SystemObject obj = dav.getDataModel().getObject(set.getString(1));
					final AttributeGroup atgObj = (AttributeGroup) dav.getDataModel().getObject(set.getLong(2));
					final short sim = set.getShort(3);
					final Blob blob = set.getBlob(4);
					long zeitpunkt = set.getLong(5);

					if (obj != null && atgObj != null && obj.isValid() && atgObj.isValid()) {
						DataWithTime dataSet = null;
						final AttributeGroup atg = atgObj;
						if (blob != null && blob.length() > 0) {
							final List<Attribut<?>> attribute = getDatenSatz(blob);
							if (attribute != null) {
								final DataWithState dataWithState = ParameterDatenFactory.getDataWithState(dav, atg,
										new Datensatz(atg.getPid(), attribute));
								if (dataWithState != null) {
									if (zeitpunkt <= 0) {
										zeitpunkt = System.currentTimeMillis();
										timeUpdates.add(new Parameter(new ParameterInfo(obj, atgObj, sim),
												new DataWithTime(dataWithState.getData(), zeitpunkt)));
									}
									dataSet = new DataWithTime(dataWithState.getData(), zeitpunkt);

									if (dataWithState.isChanged()) {
										zeitpunkt = System.currentTimeMillis();
										dataUpdates.add(new Parameter(new ParameterInfo(obj, atgObj, sim),
												new DataWithTime(dataWithState.getData(), zeitpunkt)));
									}
								}
							}
						}

						if (dataSet == null) {
							entfernbareEintrage
									.add(new DbSpezifikation(set.getString(1), set.getLong(2), set.getShort(3)));
						} else {
							synchronized (cache) {
								cache.put(obj.getId(), atgObj.getId(), sim, dataSet);
							}
						}
					} else {
						entfernbareEintrage.add(new DbSpezifikation(set.getString(1), set.getLong(2), set.getShort(3)));
					}
				}
			}
		}

		if (!timeUpdates.isEmpty()) {
			DerbyPersistenz.LOGGER.info("Aktualisiere Zeitstempel für " + timeUpdates.size() + " Einträge");

			for (final Parameter param : timeUpdates) {
				try {
					aktualisiereIntern(param.getObjekt(), param.getAtg().getId(), param.getSim(),
							param.getDataWithTime());
				} catch (final DatenSchreibenException e) {
					DerbyPersistenz.LOGGER.warning("Aktualisierung des Zeitstempels fehlgeschlagen",
							new Object[] { e.getLocalizedMessage(), param.getObjekt(), param.getAtg() });
				}
			}
		}

		if (!dataUpdates.isEmpty()) {
			DerbyPersistenz.LOGGER.info(
					"Aktualisiere Datensätze mit korrigierten Referenzen für " + dataUpdates.size() + " Einträge");

			for (final Parameter param : dataUpdates) {
				try {
					aktualisiereIntern(param.getObjekt(), param.getAtg().getId(), param.getSim(),
							param.getDataWithTime());
				} catch (final DatenSchreibenException e) {
					DerbyPersistenz.LOGGER.warning("Aktualisierung der Daten fehlgeschlagen",
							new Object[] { e.getLocalizedMessage(), param.getObjekt(), param.getAtg() });
				}
			}
		}

		DerbyPersistenz.LOGGER
				.info("Daten initialisiert: Entfernbar: " + entfernbareEintrage.size() + ": Im Cache: " + cache.size());

		DerbyPersistenz.LOGGER.info("Entferne unnötige Datenbankeinträge (" + entfernbareEintrage.size() + ")");
		for (final DbSpezifikation spez : entfernbareEintrage) {
			entferneIntern(spez.pid, spez.atg, spez.sim);
		}
		DerbyPersistenz.LOGGER.info("Unnötige Datenbankeinträge entfernt");
	}

	@Override
	public synchronized Parameter lies(final ParameterInfo parameter) {
		final Parameter result = new Parameter(parameter);
		DerbyPersistenz.LOGGER.finest("Lesen:", parameter);

		final SystemObject obj = result.getObjekt();
		final AttributeGroup atg = result.getAtg();
		final short simVariante = result.getSim();

		DataWithTime daten = null;

		if (cache.contains(obj.getId(), atg.getId(), simVariante)) {
			daten = cache.get(obj, atg, simVariante);
		}

		if (daten == null && simVariante != 0) {
			daten = cache.get(obj, atg, (short) 0);
		}

		result.setDataWithTime(daten);
		return result;
	}

	/**
	 * liest den spezifizierten Datensatz aus der Datenbank.
	 *
	 * @param obj
	 *            das Objekt
	 * @param atg
	 *            die ID der Attributgruppe
	 * @param simVariante
	 *            die Simulationsvariante
	 * @return den Datensatz oder <code>null</code>
	 */
	DataWithTime liesIntern(final SystemObject obj, final AttributeGroup atg, final short simVariante) {
		DataWithTime result = null;
		List<Attribut<?>> attribute = null;

		synchronized (lock) {

			try (PreparedStatement selectStatement = dbConnection.prepareStatement(
					"SELECT daten, zeitpunkt FROM parameterdaten WHERE pid=? AND atg=? AND simvariante=?")) {

				try (Statement s = dbConnection.createStatement()) {
					s.execute("LOCK TABLE PARAMETERDATEN IN EXCLUSIVE MODE");

					selectStatement.setString(1, obj.getPid());
					selectStatement.setLong(2, atg.getId());
					selectStatement.setInt(3, simVariante);
					try (ResultSet set = selectStatement.executeQuery()) {
						if (set.next()) {
							final Blob blob = set.getBlob(1);
							final long zeitpunkt = set.getLong(2);
							if (blob != null && blob.length() > 0) {
								attribute = getDatenSatz(blob);
								if (attribute != null) {
									result = new DataWithTime(ParameterDatenFactory.getData(dav, atg,
											new Datensatz(atg.getPid(), attribute)), zeitpunkt);
								}
							}
						}
					}
				}

				synchronized (cache) {
					cache.put(obj.getId(), atg.getId(), simVariante, result);
				}

			} catch (final SQLException e) {
				LogTools.log(DerbyPersistenz.LOGGER, ParamMessage.ErrorLesenDatensatzDerbyDb, obj, atg, simVariante,
						e.getMessage());
				reconnect();
			}
		}

		if (result != null && result.getData() != null && result.getZeitpunkt() <= 0) {
			result = new DataWithTime(result.getData(), System.currentTimeMillis());
			try {
				schreibeIntern(obj, atg.getId(), simVariante, result);
			} catch (final DatenSchreibenException e) {
				DerbyPersistenz.LOGGER.warning(
						"Zeitstempel-Aktualisierung eines Datensatzes fehlgeschlagen: " + e.getLocalizedMessage(),
						new Object[] { obj, atg, simVariante });
			}
		}
		return result;
	}

	@Override
	public synchronized void schreibe(final Parameter parameter) throws DatenSchreibenException {

		DerbyPersistenz.LOGGER.finest("Schreiben:", parameter);

		final SystemObject obj = parameter.getInfo().getObjekt();
		final AttributeGroup atg = parameter.getInfo().getAtg();

		if (DerbyPersistenz.simulationen.size() > 0) {
			// Alte Daten für Simulationen sichern
			if (parameter.getInfo().getSim() == 0) {
				DataWithTime alteDaten = null;
				if (cache.contains(obj.getId(), atg.getId(), (short) 0)) {
					alteDaten = cache.get(obj, atg, (short) 0);
				}
				for (final Short simVariante : DerbyPersistenz.simulationen) {
					if (liesIntern(obj, atg, simVariante) == null) {
						schreibeIntern(obj, atg.getId(), simVariante, alteDaten);
						synchronized (cache) {
							cache.put(obj.getId(), atg.getId(), simVariante, alteDaten);
						}
					}
				}
			}
		}

		if (parameter.getData() != null) {
			final Parameter existingData = lies(parameter.getInfo());
			if (existingData == null || !existingData.isDataAvailable()) {
				schreibeIntern(obj, atg.getId(), parameter.getSim(), parameter.getDataWithTime());
			} else {
				aktualisiereIntern(obj, atg.getId(), parameter.getSim(), parameter.getDataWithTime());
			}
			synchronized (cache) {
				cache.put(obj.getId(), atg.getId(), parameter.getSim(), parameter.getDataWithTime());
			}
		} else {
			entferne(parameter.getInfo());
		}
	}

	/**
	 * schreibt den übergebenen Datensatz in die Datenbank.
	 *
	 * @param obj
	 *            das Objekt
	 * @param atgId
	 *            die Id der Attributgruppe
	 * @param sim
	 *            die Simulationsvariante
	 * @param daten
	 *            der Datensatz
	 * @throws DatenSchreibenException
	 *             die Daten konnten nicht innerhalb der Datenbank gespeichert
	 *             werden
	 */
	private void schreibeIntern(final SystemObject obj, final long atgId, final short sim, final DataWithTime daten)
			throws DatenSchreibenException {

		final Datensatz datenSatz = (Datensatz) ParameterDatenFactory.getSerializable(dav, daten.getData());

		synchronized (lock) {

			try (PreparedStatement insertStatement = dbConnection.prepareStatement(
					"insert into parameterdaten (pid, atg, simvariante, daten, zeitpunkt) values (?, ?, ?, ?, ?)")) {

				try (Statement s = dbConnection.createStatement()) {
					s.execute("LOCK TABLE PARAMETERDATEN IN EXCLUSIVE MODE");

					insertStatement.setString(1, obj.getPid());
					insertStatement.setLong(2, atgId);
					insertStatement.setInt(3, sim);
					insertStatement.setBlob(4, getBlob(datenSatz));
					insertStatement.setLong(5, daten.getZeitpunkt());
					insertStatement.execute();
				}

			} catch (final SQLException e) {
				LogTools.log(DerbyPersistenz.LOGGER, ParamMessage.ErrorSchreibenDatensatzDerbyDb, obj,
						dav.getDataModel().getObject(atgId), sim, e.getMessage());
				reconnect();
				throw new DatenSchreibenException("Fehler beim Schreiben der Daten für: " + obj.getPid(), e);
			} catch (final IOException e) {
				LogTools.log(DerbyPersistenz.LOGGER, ParamMessage.ErrorSchreibenDatensatzDerbyDb, obj,
						dav.getDataModel().getObject(atgId), sim, e.getMessage());
				throw new DatenSchreibenException("Fehler beim Schreiben der Daten für: " + obj.getPid(), e);
			}
		}
	}

	@Override
	public void cleanup(final ParameterProvider provider) {
		Collection<de.bsvrz.puk.param.param.DerbyPersistenz.PersistenzCache.CacheKey> cachedData = null;
		synchronized (cache) {
			cachedData = new ArrayList<>(cache.daten.keySet());
		}
		for (final de.bsvrz.puk.param.param.DerbyPersistenz.PersistenzCache.CacheKey key : cachedData) {
			try {
				final SystemObject obj = dav.getDataModel().getObject(key.objId);
				if (obj != null && !(obj instanceof SystemObjectType)) {
					final AttributeGroup atg = (AttributeGroup) dav.getDataModel().getObject(key.atgId);
					final ParameterInfo info = new ParameterInfo(obj, atg, key.simVariante);
					if (!provider.isProvided(info)) {
						entferne(info);
					}
				}
			} catch (final ClassCastException e) {
				Debug.getLogger().warning(e.getLocalizedMessage(), e);
			}
		}
	}
}
