/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.simulation;

import java.util.HashSet;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.config.MutableSet;
import de.bsvrz.dav.daf.main.config.MutableSetChangeListener;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.puk.param.param.ParamMessage;
import de.bsvrz.puk.param.param.ParametrierungsModul;
import de.bsvrz.sys.funclib.bitctrl.daf.BetriebsmeldungDaten;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Realisierung des Moduls zur Verwaltung von Simulationen innerhalb der
 * Parametrierung.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public class SimulationsVerwaltung implements MutableSetChangeListener {

	/** Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/** die Parametrierung, für ide die Simulationen verwaltet werden sollen. */
	private final ParametrierungsModul parametrierung;

	/** die Menge der installierten Simulationshandler. */
	private final Set<SimulationsHandler> simulationen = new HashSet<>();

	/**
	 * erzeugt eine Instanz der Simulationsverwaltung.
	 *
	 * @param parametrierung
	 *            die verwendete Parametrierung
	 */
	public SimulationsVerwaltung(final ParametrierungsModul parametrierung) {
		SimulationsVerwaltung.LOGGER.info("Starte Submodul Simulationsverwaltung ...");
		this.parametrierung = parametrierung;

		final MutableSet set = parametrierung.getVerbindung().getLocalConfigurationAuthority()
				.getMutableSet("Simulationen");
		set.addChangeListener(this);

		for (final SystemObject simulationsObjekt : set.getElements()) {
			try {
				if (simulationsObjekt.isValid()) {
					simulationen.add(new SimulationsHandler(this, simulationsObjekt));
				} else {
					SimulationsVerwaltung.LOGGER
							.info("Ungültiges Simulationsobjekt in der Menge der Simulationen gefunden: "
									+ simulationsObjekt);
				}
			} catch (final InvalidSimulationException e) {
				LogTools.log(SimulationsVerwaltung.LOGGER, new BetriebsmeldungDaten(simulationsObjekt),
						ParamMessage.ErrorAnlegenSimulationsHandler, simulationsObjekt, e.getMessage());
			}
		}
	}

	/**
	 * setzt den Deaktivierungsstatus aller Simulationen auf INAKTIV.
	 */
	public void beendeSimulationen() {
		for (final SimulationsHandler handler : simulationen) {
			handler.aktiviereSimulation(false);
		}
	}

	/**
	 * liefert die verwendete Parametrierung.
	 *
	 * @return parametrierung
	 */
	ParametrierungsModul getParametrierung() {
		return parametrierung;
	}

	/**
	 * liefert die Datenverteilerverbindung der Simulationsverwaltung.
	 *
	 * @return die Verbindung
	 */
	ClientDavInterface getVerbindung() {
		return parametrierung.getVerbindung();
	}

	@Override
	public void update(final MutableSet set, final SystemObject[] addedObjects, final SystemObject[] removedObjects) {
		for (final SystemObject objekt : addedObjects) {
			if (objekt.isOfType("typ.simulation")) {
				try {
					if (objekt.isValid()) {
						simulationen.add(new SimulationsHandler(this, objekt));
					} else {
						SimulationsVerwaltung.LOGGER
								.info("Ungültiges Simulationsobjekt in der Menge der Simulationen gefunden: " + objekt);
					}
				} catch (final InvalidSimulationException e) {
					LogTools.log(SimulationsVerwaltung.LOGGER, new BetriebsmeldungDaten(objekt),
							ParamMessage.ErrorAnlegenSimulationsHandler, objekt, e.getMessage());
				}
			}
		}

		for (final SystemObject objekt : removedObjects) {
			if (objekt.isOfType("typ.simulation")) {
				final SimulationsHandler handler = SimulationsHandler.getHandler(objekt);
				if (handler != null) {
					handler.simulationEntfernen();
					simulationen.remove(handler);
				}
			}
		}
	}
}
